import re
from tqdm import tqdm
import tensorflow as tf

from dioptra.inference.inference_runner import InferenceRunner

class ClassifierRunner(InferenceRunner):
    def __init__(
            self, model, embeddings_layers,
            logits_layer, class_names,
            metadata=None):
        """
        Utility to perform model inference on a dataset and extract layers needed for AL.

        Parameters:
            model: model to be used to inference
            embeddings_layers: an array of layer names that should be used as embeddings. Can be a jq style path to an embedding layer like [0].my_embedding
            logits_layer: the name of the logit layer (pre softmax) to be used for AL. Can be a jq style path to an embedding layer like [0].my_logits
            class_names: the class names corresponding to each logit. Indexes should match the logit layer
            metadata: a list of dioptra style metadata to be added to teh datapoints. The indexes in this list should match the indexes in the dataset
        """

        super().__init__()

        self.model = model
        self.embeddings_layers = embeddings_layers
        self.logits_layer = logits_layer
        self.class_names = class_names
        self.metadata = metadata

        input_layer = model.inputs
        if input_layer is None:
            input_layer = model.layers[0].inputs

        output_layers = {}

        for layer in embeddings_layers + [logits_layer]:
            output_layers[layer] = self._get_layer_by_name(layer).output

        self.logging_model = tf.keras.Model(
            inputs=input_layer,
            outputs=output_layers)

    def _get_layer_by_name(self, name):
        split = name.split('.')
        current_layer = self.model
        for part in split:
            if re.match('\[[0-9]+\]', part):
                index = int(part.replace('[', '').replace(']', ''))
                current_layer = current_layer.layers[index]
            else:
                current_layer = current_layer.get_layer(part)
        return current_layer


    def run(self, dataset):
        """
        Run the inference on a dataset and upload results to dioptra

        Parameters:
            dataset: a tf.data.Dataset
                Should be batched and pre processed to only return the data, not the groundtruth
                Should not be shuffled if used with a metadata list
        """

        records = []

        global_idx = 0

        for batch_index, batch in tqdm(enumerate(dataset), desc='running inference...'):
            output = self.logging_model(batch)
            for batch_idx in range(batch.shape[0]):
                records.extend(self._build_records(batch_idx, global_idx, output))
                global_idx += 1
                if len(records) > self.max_batch_size:
                    self._ingest_data(records)
                    records = []
        if len(records) > 0:
            self._ingest_data(records)
            records = []

    def _build_records(self, record_batch_idx, record_global_idx, output):

        my_record = {
            **({
                'prediction': {
                    'logits': output[self.logits_layer][record_batch_idx].numpy(),
                    'class_names': self.class_names
                }
               } if self.logits_layer is not None and self.class_names is not None else {}
            ),
            'model_type': 'CLASSIFIER',
            **(self.metadata[record_global_idx] \
               if self.metadata and len(self.metadata) > record_global_idx else {}
            )
        }

        my_records = []

        for my_layer in self.embeddings_layers:
            record_tags = my_record.get('tags', {})
            if 'embeddings_name' not in record_tags:
                record_tags['embeddings_name'] = my_layer
                my_record['tags'] = record_tags
            layer_record = {
                'embeddings': output[my_layer][record_batch_idx].numpy(),
                **my_record
            }
            my_records.append(layer_record)

        if len(my_records) == 0:
            my_records.append(my_record)

        return my_records



from dioptra.miners.base_miner import BaseMiner
from dioptra.lake.utils import _list_miner_metadata

def list_miners():
    """
    List all the miners

    """

    miner_list = []

    for metadata in _list_miner_metadata():
        my_miner = BaseMiner()
        my_miner.miner_id = metadata['_id']
        my_miner.miner_name = metadata['display_name']
        miner_list.append(my_miner)

    return miner_list



"""
Utility module

"""

import datetime

from dioptra.schemas import (
    object_detection_schema,
    multiple_object_tracking_schema,
    feature_schema,
    tag_schema,
    embeddings_schema,
    logits_schema,
    image_metadata_schema,
    text_metadata_schema,
    audio_metadata_schema,
    video_metadata_schema,
    question_answering_groundtruth_schema,
    question_answering_prediction_schema,
    automated_speech_recogniton_schema,
    auto_completion_groundtruth_schema,
    auto_completion_prediction_schema,
    semantic_similarity_schema,
    semantic_similarity_input_schema,
    classification_schema
)

from dioptra.supported_types import ModelTypes, InputTypes

from dioptra.supported_types import SupportedTypes

def validate_model_type(model_type):
    return isinstance(model_type, SupportedTypes)

def validate_timestamp(timestamp):
    return isinstance(timestamp, datetime.datetime)

def validate_tags(tags):
    return tag_schema().is_valid(tags)

def validate_text(text):
    return isinstance(text, str)

def validate_input_data(input_data, model_type):
    if model_type.input_type == InputTypes.PAIRED_TEXT:
        return semantic_similarity_input_schema().is_valid(input_data)
    return False

def validate_annotations(annotations, model_type):

    if model_type.model_type == ModelTypes.CLASSIFIER:
        if isinstance(annotations, str):
            return True
        elif classification_schema().is_valid(annotations):
            return True
        else:
            return False
    elif model_type.model_type == ModelTypes.OBJECT_DETECTION:
        return object_detection_schema().is_valid(annotations)
    elif model_type.model_type == ModelTypes.MULTIPLE_OBJECT_TRACKING:
        return multiple_object_tracking_schema().is_valid(annotations)
    elif model_type.model_type == ModelTypes.QUESTION_ANSWERING:
        if question_answering_groundtruth_schema().is_valid(annotations):
            return True
        elif question_answering_prediction_schema().is_valid(annotations):
            return True
        return False
    elif model_type.model_type == ModelTypes.AUTOMATED_SPEECH_RECOGNITION:
        return automated_speech_recogniton_schema().is_valid(annotations)
    elif model_type.model_type == ModelTypes.AUTO_COMPLETION:
        if auto_completion_groundtruth_schema().is_valid(annotations):
            return True
        elif auto_completion_prediction_schema().is_valid(annotations):
            return True
        return False
    elif model_type.model_type == ModelTypes.SEMANTIC_SIMILARITY:
        return semantic_similarity_schema().is_valid(annotations)
    else:
        return False

def validate_confidence(confidence):
    return isinstance(confidence, float)

def validate_features(features):
    return feature_schema().is_valid(features)

def validate_embeddings(embeddings):
    return embeddings_schema().is_valid(embeddings)

def validate_logits(embeddings):
    return embeddings_schema().is_valid(embeddings)

def validate_image_metadata(image_metadata):
    return image_metadata_schema().is_valid(image_metadata)

def validate_text_metadata(text_metadata):
    return text_metadata_schema().is_valid(text_metadata)

def validate_audio_metadata(audio_metadata):
    return audio_metadata_schema().is_valid(audio_metadata)

def validate_video_metadata(video_metadata):
    return video_metadata_schema().is_valid(video_metadata)



import re
from tqdm import tqdm
import torch

from dioptra.inference.inference_runner import InferenceRunner

class ClassifierRunner(InferenceRunner):
    def __init__(
            self, model, embeddings_layers=[],
            logits_layer=None, class_names=[],
            metadata=None,
            data_transform=None,
            device='cpu'):
        """
        Utility to perform model inference on a dataset and extract layers needed for AL.

        Parameters:
            model: model to be used to inference
            embeddings_layers: an array of layer names that should be used as embeddings
            logits_layer: the name of the logit layer (pre softmax) to be used for AL
            class_names: the class names corresponding to each logit. Indexes should match the logit layer
            metadata: a list of dioptra style metadata to be added to teh datapoints. The indexes in this list should match the indexes in the dataset
            data_transform: a transform function that will be called before the model is called. Should only return the data, without the groundtruth
            device: the devide to be use to perform the inference
        """

        super().__init__()

        self.model = model
        self.embeddings_layers = embeddings_layers
        self.logits_layer = logits_layer
        self.class_names = class_names
        self.metadata = metadata
        self.data_transform = data_transform
        self.device = device

        self.activation = {}

        for my_layer_name in embeddings_layers + [logits_layer]:
            if my_layer_name is not None:
                my_layer = self._get_layer_by_name(my_layer_name)
                my_layer.register_forward_hook(self._get_activation(my_layer_name))

    def _get_layer_by_name(self, name):
        split = name.split('.')
        current_layer = self.model
        for part in split:
            if re.match('\[[0-9]+\]', part):
                index = int(part.replace('[', '').replace(']', ''))
                current_layer = current_layer[index]
            else:
                current_layer = getattr(current_layer, part)
        return current_layer

    def _get_activation(self, name):
        def hook(model, input, output):
            if hasattr(output, 'last_hidden_state'):
                self.activation[name] = output.last_hidden_state
            else:
                self.activation[name] = output.detach()
        return hook

    def run(self, dataloader):
        """
        Run the inference on a dataset and upload results to dioptra

        Parameters:
            dataset: a torch.utils.data.Dataset
                Should be batched. data_transform can be used to pre process teh data to only return the data, not the groundtruth
                Should not be shuffled if used with a metadata list
        """

        self.model.eval()
        self.model.to(self.device)

        records = []

        global_idx = 0
        if hasattr(dataloader, 'dataset'):
            dataset_size = len(dataloader.dataset) # we are using a dataloader
        else:
            dataset_size = len(dataloader)  # we are using a dataset directly
        for batch_index, batch in tqdm(enumerate(dataloader), desc='running inference...'):
            if self.data_transform:
                batch = self.data_transform(batch)
            batch = batch.to(self.device)
            with torch.no_grad():
                self.model(batch)
            for batch_idx, _ in enumerate(batch):
                records.extend(self._build_records(batch_idx, global_idx))
                global_idx += 1
                if len(records) > self.max_batch_size:
                    self._ingest_data(records)
                    records = []
                if global_idx > dataset_size:
                    break
            if global_idx > dataset_size:
                break
        if len(records) > 0:
            self._ingest_data(records)
            records = []

    def _build_records(self, record_batch_idx, record_global_idx):

        my_record = {
            **({
                'prediction': {
                    'logits': self.activation[self.logits_layer][record_batch_idx].cpu().numpy(),
                    'class_names': self.class_names
                }
               } if self.logits_layer in self.activation  and self.class_names else {}
            ),
            'model_type': 'CLASSIFIER',
            **(self.metadata[record_global_idx] \
               if self.metadata and len(self.metadata) > record_global_idx else {}
            )
        }

        my_records = []

        for my_layer in self.embeddings_layers:
            if my_layer not in self.activation:
                continue
            record_tags = my_record.get('tags', {})
            if 'embeddings_name' not in record_tags:
                record_tags['embeddings_name'] = my_layer
                my_record['tags'] = record_tags
            layer_record = {
                'embeddings': self.activation[my_layer][record_batch_idx].cpu().numpy(),
                **my_record
            }
            my_records.append(layer_record)

        if len(my_records) == 0:
            my_records.append(my_record)

        return my_records



import os
import io
import requests
import base64
import time
from multiprocessing import Pool
import io
import uuid
from datetime import datetime

import smart_open
import PIL
import pandas as pd
import numpy as np
import lz4
import boto3
from botocore.exceptions import ClientError
import tqdm
import orjson
import mgzip

# TODO: figure out how to return status codes form lambda functions.
# See the wip_return_error_code branch of the infrastructure repo.
def _raise_for_apigateway_errormessage(response):
    if response is not None and 'errorMessage' in response:
        raise RuntimeError(response['errorMessage'])

def query_dioptra_app(method, path, body=None):
    api_key = os.environ.get('DIOPTRA_API_KEY', None)
    if api_key is None:
        raise RuntimeError('DIOPTRA_API_KEY env var is not set')

    app_endpoint = os.environ.get('DIOPTRA_APP_ENDPOINT', 'https://app.dioptra.ai')

    r = getattr(requests, method.lower())(f'{app_endpoint}{path}', headers={
        'content-type': 'application/json',
        'x-api-key': api_key
    }, json=body)
    r.raise_for_status()

    return r.json()

def download_from_lake(filters, limit=None, order_by=None, desc=None, fields=['*']):
    """
    Downloading metadata from the data lake

    Parameters:
        filters: dioptra style filters to select the data to be queried from
        limit: limit to selected the data
        order_by: field to use to sort the data to control how limit is performed
        desc: whether to order by dec or not
        fields: array of fields to be queried. By default all fields are queried
    """

    return pd.json_normalize(query_dioptra_app(
        'POST',
        '/api/metrics/select',
        {
            'select': ','.join(fields),
            'filters': filters,
            **({'limit': limit} if limit is not None else {}),
            **({'order_by': order_by} if order_by is not None else {}),
            **({'desc': desc} if desc is not None else {})
        }
    ))

def select_datapoints(filters, limit=None, order_by=None, desc=None, fields=['*']):
    """
    Select metadata from the data lake

    Parameters:
        filters: dioptra style filters to select the data to be queried from
        limit: limit to selected the data
        order_by: field to use to sort the data to control how limit is performed
        desc: whether to order by dec or not
        fields: array of fields to be queried. By default all fields are queried
    """

    return pd.DataFrame(query_dioptra_app(
        'POST',
        '/api/datapoints/select',
        {
            'selectColumns': fields,
            'filters': filters,
            **({'limit': limit} if limit is not None else {}),
            **({'orderBy': order_by} if order_by is not None else {}),
            **({'desc': desc} if desc is not None else {})
        }
    ))

def get_predictions(datapoint_ids):
    """
    Get predictions for a set of datapoints

    Parameters:
        datapoint_ids: list of datapoint ids to get predictions for
    """

    return pd.json_normalize(query_dioptra_app(
        'POST',
        '/api/predictions/get',
        {
            'datapointIds': datapoint_ids
        }
    ))

def delete_predictions(prediction_ids):
    """
    Delete predictions from the data lake

    Parameters:
        prediction_ids: list of prediction ids to delete
    """

    return query_dioptra_app(
        'POST',
        '/api/predictions/delete',
        {
            'predictionIds': prediction_ids
        }
    )

def get_groundtruths(datapoint_ids):
    """
    Get groundtruths for a set of datapoints

    Parameters:
        datapoint_ids: list of datapoint ids to get groundtruths for
    """

    return pd.json_normalize(query_dioptra_app(
        'POST',
        '/api/groundtruths/get',
        {
            'datapointIds': datapoint_ids
        }
    ))

def delete_groundtruths(groundtruth_ids):
    """
    Delete groundtruths from the data lake

    Parameters:
        groundtruth_ids: list of groundtruth ids to delete
    """

    return query_dioptra_app(
        'POST',
        '/api/groundtruths/delete',
        {
            'groundtruthIds': groundtruth_ids
        }
    )

def delete_datapoints(filters, limit=None, order_by=None, desc=None):
    """
    Delete metadata from the data lake

    Parameters:
        filters: dioptra style filters to select the data to be deleted
        limit: limit to selected the data
        order_by: field to use to sort the data to control how limit is performed
        desc: whether to order by dec or not
    """

    return query_dioptra_app(
        'POST',
        '/api/datapoints/delete',
        {
            'filters': filters
        }
    )

def delete_from_lake(filters, limit=None, order_by=None, desc=None):
    """
    Delete metadata from the data lake

    Parameters:
        filters: dioptra style filters to select the data to be deleted
        limit: limit to selected the data
        order_by: field to use to sort the data to control how limit is performed
        desc: whether to order by dec or not
    """

    return query_dioptra_app(
        'POST',
        '/api/metrics/delete',
        {
            'filters': filters,
            **({'limit': limit} if limit is not None else {}),
            **({'order_by': order_by} if order_by is not None else {}),
            **({'desc': desc} if desc is not None else {})
        }
    )

def upload_to_lake(records):
    """
    Uploading metadata to the data lake

    Parameters:
        records: array of dipotra style records. See teh doc for accepted formats
    """

    api_key = os.environ.get('DIOPTRA_API_KEY', None)
    if api_key is None:
        raise RuntimeError('DIOPTRA_API_KEY env var is not set')

    api_endpoint = os.environ.get('DIOPTRA_API_ENDPOINT', 'https://api.dioptra.ai/events')

    try:
        r = requests.post(api_endpoint, headers={
            'content-type': 'application/json',
            'x-api-key': api_key
        }, json={
            'records': records
        })
        r.raise_for_status()
        response = r.json()
        _raise_for_apigateway_errormessage(response)
    except requests.exceptions.RequestException as err:
        print('There was an error uploading to the lake ...')
        raise err

    return response

def upload_to_lake_from_s3(bucket_name, object_name, storage_type='s3'):
    """
    Uploading metadata to the data lake from an s3 bucket
    The data should be a new line delimited JSON and can be compressed with Gzip
    Will generate a signed URL and pass it to the dioptra ingestion
    boto3 credentials should be configured

    Parameters:
        bucket_name: name of the bucket where the records are
        object_name: name of the key in the bucket where the records are
    """

    api_key = os.environ.get('DIOPTRA_API_KEY', None)
    if api_key is None:
        raise RuntimeError('DIOPTRA_API_KEY env var is not set')

    api_endpoint = os.environ.get('DIOPTRA_API_ENDPOINT', 'https://api.dioptra.ai/events')

    if storage_type == 's3':
        signed_url = _generate_s3_signed_url(bucket_name, object_name)

    try:
        r = requests.post(api_endpoint, headers={
            'content-type': 'application/json',
            'x-api-key': api_key
        }, json={
            'urls': [signed_url]
        })
        r.raise_for_status()
        response = r.json()
        _raise_for_apigateway_errormessage(response)
    except requests.exceptions.RequestException as err:
        print('There was an error uploading to the lake ...')
        raise err

    return response

def wait_for_upload(upload_id):
    """
    Wait for the upload status to be SUCCEEDED | FAILED | TIMED_OUT | ABORTED
    """
    if upload_id is None:
        raise RuntimeError('upload_id must not be None')

    api_key = os.environ.get('DIOPTRA_API_KEY', None)
    if api_key is None:
        raise RuntimeError('DIOPTRA_API_KEY env var is not set')

    app_endpoint = os.environ.get('DIOPTRA_APP_ENDPOINT', 'https://app.dioptra.ai')
    sleepTimeSecs = 1
    totalSleepTimeSecs = 0
    try:
        while True:
            if totalSleepTimeSecs > 900:
                raise RuntimeError('Timed out waiting for the upload to finish.')

            r = requests.get(f'{app_endpoint}/api/ingestion/executions/{upload_id}', headers={
                'content-type': 'application/json',
                'x-api-key': api_key
            })
            r.raise_for_status()
            upload = r.json()
            if upload['status'] in ['SUCCEEDED', 'FAILED', 'TIMED_OUT', 'ABORTED']:
                return upload
            else:
                time.sleep(sleepTimeSecs)
                totalSleepTimeSecs += sleepTimeSecs
                sleepTimeSecs = min(sleepTimeSecs * 2, 60)
    except requests.exceptions.RequestException as err:
        print('There was an error waiting for the upload to finish ...')
        raise err

def _generate_s3_signed_url(bucket_name, object_name):

    s3_client = boto3.client('s3')
    try:
        response = s3_client.generate_presigned_url(
            'get_object', Params={'Bucket': bucket_name, 'Key': object_name}, ExpiresIn=3600)
    except ClientError as err:
        print('There was an error getting a signed URL ...')
        raise err

    return response

def _list_dataset_metadata():
    """
    List all the dataset metadata

    """

    return query_dioptra_app('GET', '/api/dataset')

def _list_miner_metadata():
    """
    List all the miner metadata

    """

    return query_dioptra_app('GET', '/api/miner')


def _encode_np_array(np_array):
    """
    Encode and compress a np array

    Parameters:
        np_array: the np array to be encoded

    """
    if not isinstance(np_array, np.ndarray):
        raise RuntimeError('Can only encode numpy arrays')

    bytes_buffer = io.BytesIO()
    np.save(bytes_buffer, np_array)

    return base64.b64encode(
        lz4.frame.compress(
            bytes_buffer.getvalue(),
            compression_level=lz4.frame.COMPRESSIONLEVEL_MAX
        )).decode('ascii')

def _decode_to_np_array(value):
    """
    Decode a compress a np array

    Parameters:
        value: the string containing the np array

    """
    decoded_bytes = lz4.frame.decompress(base64.b64decode(value))
    return np.load(io.BytesIO(decoded_bytes), allow_pickle=True)


def _format_groundtruth(groundtruth, gt_type, class_names=None):
    """
    Utility formatting the groundtruth field according to the model type

    Parameters:
        groundtruth: the groundtruth field
        gt_type: the type of gt. Supported types CLASSIFIER, SEMANTIC_SEGMENTATION
        class_names: a list of class names. If the groundtruth contains indexes, it will be used to convert them to names

    """
    if gt_type == 'CLASSIFIER':
        if isinstance(groundtruth, int) and class_names is not None:
            class_name = class_names[groundtruth]
        else:
            class_name = groundtruth
        return {
            'class_name': class_name
        }
    if gt_type == 'SEMANTIC_SEGMENTATION':
        if isinstance(groundtruth, PIL.Image.Image):
            gt_array = np.array(groundtruth)
        if isinstance(groundtruth, np.ndarray):
            gt_array = groundtruth
        if isinstance(groundtruth, list):
            gt_array = np.array(groundtruth)
        return {
            'encoded_segmentation_class_mask': _encode_np_array(gt_array),
            **({'class_names': class_names} if class_names is not None else {})
        }


def upload_image_dataset(
    dataset, dataset_type,
    image_field=None,
    groundtruth_field=None,
    datapoints_metadata=None,
    class_names=None, dataset_tags=None,
    max_batch_size=1000, num_workers=1):
    """
    Upload an image dataset to Dioptra.
    The images will be uploaded to a bucket specified with DIOPTRA_UPLOAD_BUCKET.
    The metadata will be uploaded to Dioptra and point to this bucket

    Parameters:
        dataset: the dataset to upload. Should be iteratable.
        image_field: the name of the field containing the image. Should be in PIL format
        dataset_type: the type of data, Supported CLASSIFIER, SEMANTIC_SEGMENTATION
        groundtruth_field: the name of the field containing the groundtruth
        datapoints_metadata:
            a list of metadata to be added to each datapoint.
            should already be formatted.
            if it contains a 'image_id' field, this will be used to name the images in s3
        class_names: a list of class names to convert indexes in the groundtruth to class names
        class_names: a dict of tags to be added to the entire dataset
        max_batch_size: the maximum batch uplodd size
        num_workers: number of parallel workers to upload the images to the bucket

    """


    def _img_upload(payload):
        image_bytes, img_url = payload[0], payload[1]
        with smart_open.open(img_url, 'wb') as file:
            file.write(image_bytes)

    def _upload_data(storage_type, bucket, prefix, dataset_metadata, img_payload, num_workers):
        with Pool(num_workers) as my_pool:
            my_pool.map(_img_upload, img_payload)

        file_name = os.path.join(
            prefix, 'logs', f'{str(uuid.uuid4())}_{datetime.utcnow().isoformat()}.ndjson')
        logs_url = f'{storage_type}://{os.path.join(bucket, file_name)}'

        payload = b'\n'.join(
            [orjson.dumps(record, option=orjson.OPT_SERIALIZE_NUMPY)for record in dataset_metadata])
        compressed_payload = mgzip.compress(payload, compresslevel=2)

        with smart_open.open(logs_url, 'wb') as file:
            file.write(compressed_payload)

        return upload_to_lake_from_s3(bucket, file_name, storage_type)

    def _build_img_url(storage_type, bucket, prefix, image_id, pil_image):
        img_format = pil_image.format
        root, ext = os.path.splitext(image_id)
        if not ext:
            image_id += f'.{img_format.lower()}'
        return f'{storage_type}://{os.path.join(bucket, prefix, "images", image_id)}'

    api_key = os.environ.get('DIOPTRA_API_KEY', None)
    if api_key is None:
        raise RuntimeError('DIOPTRA_API_KEY env var is not set')

    s3_bucket = os.environ.get('DIOPTRA_UPLOAD_BUCKET', None)

    if s3_bucket is None:
        raise RuntimeError('DIOPTRA_UPLOAD_BUCKET env var is not set')

    s3_prefix_bucket = os.environ.get('DIOPTRA_UPLOAD_PREFIX', '')
    storage_type = os.environ.get('DIOPTRA_UPLOAD_BUCKET_TYPE', 's3')

    dataset_metadata = []
    img_payload = []
    upload_ids = []

    for index, row in tqdm.tqdm(enumerate(dataset), desc='uploading dataset...'):

        if image_field is not None and image_field not in row:
            raise RuntimeError(f'{image_field} is not in the dataset')
        if groundtruth_field is not None and groundtruth_field not in row:
            raise RuntimeError(f'{groundtruth_field} is not in the dataset')

        if image_field is not None:
            pil_img = row[image_field]
            if not isinstance(pil_img, PIL.Image.Image):
                raise Exception('the image provided was not a PIL image')
            img_width, img_height = pil_img.size
            image_id = str(index)

        my_metadata = None
        if datapoints_metadata is not None:
            my_metadata = datapoints_metadata[index]
            if 'image_id' in my_metadata:
                image_id = my_metadata['image_id']

        img_url = _build_img_url(storage_type, s3_bucket, s3_prefix_bucket, image_id, pil_img)

        in_mem_file = io.BytesIO()
        pil_img.save(in_mem_file, format=pil_img.format)

        datapoint_tags = {}
        image_metadata = {}

        if  my_metadata is not None and 'tags' in my_metadata:
            datapoint_tags.update(my_metadata['tags'])
        if dataset_tags is not None:
            datapoint_tags.update(dataset_tags)
        if  my_metadata is not None and 'image_metadata' in my_metadata:
            image_metadata.update(my_metadata['image_metadata'])

        image_metadata.update({
            'uri': img_url,
            'width': img_width,
            'height': img_height
        })

        img_payload.append((in_mem_file.getvalue(), img_url))
        dataset_metadata.append({
            'image_metadata': image_metadata,
            'model_type': dataset_type,
            **({
                'groundtruth': _format_groundtruth(row[groundtruth_field], dataset_type, class_names)
            } if groundtruth_field is not None else {}),
            **(my_metadata if my_metadata is not None else {}),
            **({'tags': datapoint_tags} if len(datapoint_tags) > 0 else {})
        })

        if len(img_payload) > max_batch_size:
            upload_ids.append(_upload_data(
                storage_type, s3_bucket, s3_prefix_bucket, dataset_metadata, img_payload, num_workers))
            dataset_metadata = []
            img_payload = []

    if len(img_payload) > 0:
        upload_ids.append(_upload_data(
            storage_type, s3_bucket, s3_prefix_bucket, dataset_metadata, img_payload, num_workers))

    return upload_ids



import os
import sys
import fire

from dioptra.api import Logger

API_KEY = os.environ.get('DIOPTRA_API_KEY')
DIOPTRA_BATCH_BUCKET = os.environ.get('DIOPTRA_BATCH_BUCKET', 'dioptra-batch-input-prod')

def upload_batch(ndjson_file, validate_sample=0.01):
    """Upload a batch of data to Dioptra.
    :param str ndjson_file: The path of the batch file in newline-delimited json (.ndjson) format.
    :param float validate_sample: Validate a random sample of the data (between 0-1) 
    """
    if API_KEY is None:
        print('No DIOPTRA_API_KEY found in the environment')
        sys.exit(1)

    dioptra_logger = Logger(api_key=API_KEY)
    dioptra_logger.commit_ndjson_file(
        ndjson_file,
        validate_sample=validate_sample,
        batch_bucket=DIOPTRA_BATCH_BUCKET
    )

if __name__ == '__main__':
    fire.Fire({
        'upload_batch': upload_batch
    }, name='dioptra')



from enum import Enum


class ModelTypes(Enum):
    CLASSIFIER = 'CLASSIFIER'
    OBJECT_DETECTION = 'OBJECT_DETECTION'
    QUESTION_ANSWERING = 'Q_N_A'
    AUTOMATED_SPEECH_RECOGNITION = 'ASR'
    AUTO_COMPLETION = 'AUTO_COMPLETION'
    SEMANTIC_SIMILARITY = 'SEMANTIC_SIMILARITY'
    MULTIPLE_OBJECT_TRACKING = 'MULTIPLE_OBJECT_TRACKING'

class InputTypes(Enum):
    TABULAR = 'TABULAR'
    IMAGE = 'IMAGE'
    TEXT = 'TEXT'
    PAIRED_TEXT = 'PAIRED_TEXT'
    AUDIO = 'AUDIO'
    VIDEO = 'VIDEO'

class SupportedTypes(Enum):
    TABULAR_CLASSIFIER = 'TABULAR_CLASSIFIER', ModelTypes.CLASSIFIER, InputTypes.TABULAR
    IMAGE_CLASSIFIER = 'IMAGE_CLASSIFIER', ModelTypes.CLASSIFIER, InputTypes.IMAGE
    TEXT_CLASSIFIER = 'TEXT_CLASSIFIER', ModelTypes.CLASSIFIER, InputTypes.TEXT
    OBJECT_DETECTION = 'OBJECT_DETECTION', ModelTypes.OBJECT_DETECTION, InputTypes.IMAGE
    QUESTION_ANSWERING = 'QUESTION_ANSWERING', ModelTypes.QUESTION_ANSWERING, InputTypes.TEXT
    AUTOMATED_SPEECH_RECOGNITION = 'AUTOMATED_SPEECH_RECOGNITION', \
        ModelTypes.AUTOMATED_SPEECH_RECOGNITION, InputTypes.AUDIO
    AUTO_COMPLETION = 'AUTO_COMPLETION', ModelTypes.AUTO_COMPLETION, InputTypes.TEXT
    SEMANTIC_SIMILARITY = 'SEMANTIC_SIMILARITY', \
        ModelTypes.SEMANTIC_SIMILARITY, InputTypes.PAIRED_TEXT
    MULTIPLE_OBJECT_TRACKING = 'MULTIPLE_OBJECT_TRACKING', ModelTypes.MULTIPLE_OBJECT_TRACKING, InputTypes.VIDEO

    def __new__(cls, *args, **kwds):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __init__(self, _, model_type, input_type):
        self._model_type = model_type
        self._input_type = input_type

    def __str__(self):
        return self.value

    @property
    def model_type(self):
        return self._model_type
    @property
    def input_type(self):
        return self._input_type



import logging
import datetime
import dateutil.parser
import orjson
import os
import gzip
import threading
import random
import traceback
from smart_open import open as smart_open

from tqdm import tqdm
import boto3
from boto3.s3.transfer import TransferConfig
from botocore import UNSIGNED
from botocore.client import Config
import uuid

import dioptra
from dioptra.client import Client
from dioptra.supported_types import SupportedTypes
from dioptra.utils import (
    validate_tags,
    validate_embeddings,
    validate_features,
    validate_annotations,
    validate_confidence,
    validate_image_metadata,
    validate_timestamp,
    validate_text_metadata,
    validate_text,
    validate_audio_metadata,
    validate_video_metadata,
    validate_model_type,
    validate_input_data
)

s3_client = boto3.client('s3', config=Config(signature_version=UNSIGNED))

class FileProgress(object):
    def __init__(self, filenames):
        self._size = 0

        for filename in filenames:
            self._size += float(os.path.getsize(filename))

        self._lock = threading.Lock()
        self._pbar = tqdm(
            total=self._size,
            unit='B',
            desc='Uploading to Dioptra'
        )

    def __call__(self, bytes_amount):
        with self._lock:
            self._pbar.update(bytes_amount)

BATCH_INGEST_MAX_SIZE = 5000

class Logger:
    def __init__(
        self,
        api_key,
        endpoint_url='https://api.dioptra.ai',
        synchronous_mode=False,
        max_workers=5,
        batch_size=20,
        queue_size=100
    ):
        self.api_key = api_key
        self.endpoint_url = endpoint_url
        self.event_url = endpoint_url + '/events'
        self._headers = {
            'x-api-key': api_key,
            'Content-Type': 'application/json'
        }
        # TODO: reenable setting synchronous_mode
        self.client = Client(max_workers, queue_size, synchronous_mode=True)
        self.batch = []
        self.batch_size = batch_size

    def log(
        self,
        request_id,
        timestamp=None,
        dataset_id=None,
        benchmark_id=None,
        model_id=None,
        model_version=None,
        model_type=None,
        groundtruth=None,
        prediction=None,
        confidence=None,
        features=None,
        embeddings=None,
        input_data=None,
        image_metadata=None,
        text=None,
        text_metadata=None,
        tags=None,
        audio_metadata=None,
        validate_sample=True
    ):
        """
        The log interface sends a set of events to the dioptra api endpoint.
        Yet, the events are not commited yet.
        As long as a event is not commited, it won't be available for analysis.
        Once a committed event with the same `request_id` is sent,
        it will be aggregated with previous events in a session that share the same `request_id`
        A session is 1 day long.

        """

        payload = self.package_payload(
            model_id,
            dataset_id,
            benchmark_id,
            model_version,
            model_type,
            timestamp,
            request_id,
            groundtruth,
            prediction,
            confidence,
            features,
            embeddings,
            input_data,
            image_metadata,
            text,
            text_metadata,
            tags,
            audio_metadata,
            False,
            validate_sample=1 if validate_sample else 0
        )

        self.client.call(
            payload=payload,
            endpoint_url=self.event_url,
            headers=self._headers)

    def commit(
        self,
        request_id,
        timestamp=None,
        model_id=None,
        dataset_id=None,
        benchmark_id=None,
        model_version=None,
        model_type=None,
        groundtruth=None,
        prediction=None,
        confidence=None,
        features=None,
        embeddings=None,
        input_data=None,
        image_metadata=None,
        text=None,
        text_metadata=None,
        tags=None,
        audio_metadata=None,
        validate_sample=True
    ):
        """
        The commit interface sends a set of events to the dioptra api endpoint and commits them.
        A committed event triggers the aggregation of all previous events with the same `request_id`
        in the same session. A seesion is 1 day long.
        Once a committed all events are available for analysis in the dioptra UI.

        """

        payload = self.package_payload(
            model_id,
            dataset_id,
            benchmark_id,
            model_version,
            model_type,
            timestamp,
            request_id,
            groundtruth,
            prediction,
            confidence,
            features,
            embeddings,
            input_data,
            image_metadata,
            text,
            text_metadata,
            tags,
            audio_metadata,
            True,
            validate_sample=1 if validate_sample else 0
        )

        self.client.call(
            payload=payload,
            endpoint_url=self.event_url,
            headers=self._headers)

    def add_to_batch_log(
        self,
        request_id,
        timestamp=None,
        dataset_id=None,
        benchmark_id=None,
        model_id=None,
        model_version=None,
        model_type=None,
        groundtruth=None,
        prediction=None,
        confidence=None,
        features=None,
        embeddings=None,
        input_data=None,
        image_metadata=None,
        text=None,
        text_metadata=None,
        tags=None,
        audio_metadata=None,
        validate_sample=True
    ):

        payload = self.package_payload(
            model_id,
            dataset_id,
            benchmark_id,
            model_version,
            model_type,
            timestamp,
            request_id,
            groundtruth,
            prediction,
            confidence,
            features,
            embeddings,
            input_data,
            image_metadata,
            text,
            text_metadata,
            tags,
            audio_metadata,
            False,
            validate_sample=1 if validate_sample else 0
        )

        self.batch.extend(payload)


    def add_to_batch_commit(
        self,
        request_id,
        timestamp=None,
        dataset_id=None,
        benchmark_id=None,
        model_id=None,
        model_version=None,
        model_type=None,
        groundtruth=None,
        prediction=None,
        confidence=None,
        features=None,
        embeddings=None,
        input_data=None,
        image_metadata=None,
        text=None,
        text_metadata=None,
        tags=None,
        audio_metadata=None,
        validate_sample=True
    ):

        payload = self.package_payload(
            model_id,
            dataset_id,
            benchmark_id,
            model_version,
            model_type,
            timestamp,
            request_id,
            groundtruth,
            prediction,
            confidence,
            features,
            embeddings,
            input_data,
            image_metadata,
            text,
            text_metadata,
            tags,
            audio_metadata,
            True,
            validate_sample=1 if validate_sample else 0
        )

        self.batch.extend(payload)

    def submit_batch(self):
        batches = [
            self.batch[i:i + self.batch_size] for i in range(0, len(self.batch), self.batch_size)]
        for batch in tqdm(batches, desc='Batch upload progress'):
            self.client.call(
                payload=batch,
                endpoint_url=self.event_url,
                headers=self._headers)


    def commit_ndjson_file(self, ndjson_filename, validate_sample=0.01, batch_bucket='dioptra-batch-input-prod'):
        """
        Loading the config from file
        """
        events_batch = []
        progress = tqdm(
            unit='Event',
            desc='Uploading to Dioptra'
        )
        def flush_batch(events_batch, progress):
            batch_key = str(uuid.uuid4()) + '.ndjson.gz'
            batch_tmp_filepath = f'/tmp/dioptra/{batch_key}'

            try:
                events_payload = b'\n'.join([orjson.dumps(e) for e in events_batch])
                os.makedirs(os.path.dirname(batch_tmp_filepath), exist_ok=True)
                with open(batch_tmp_filepath, 'wb') as tmp_file:
                    tmp_file.write(gzip.compress(events_payload))

                GB = 1024 ** 3
                transfer = boto3.s3.transfer.S3Transfer(s3_client, config=TransferConfig(multipart_threshold=5*GB))
                transfer.upload_file(
                    batch_tmp_filepath, batch_bucket, batch_key,
                    extra_args={'Metadata': {'x-api-key': self.api_key}}
                )
                progress.update(len(events_batch))
            except Exception as e:
                print(traceback.format_exc())
            finally:
                os.remove(batch_tmp_filepath)

        for _ in range(1):
            for line in smart_open(ndjson_filename):
                try:
                    parsed_line = orjson.loads(line)
                except:
                    raise Exception('Failed to parse json datapoint: ' + str(line))

                events_batch.extend(self.package_payload(**parsed_line, validate_sample=validate_sample))

                if len(events_batch) >= BATCH_INGEST_MAX_SIZE:
                    flush_batch(events_batch, progress)
                    events_batch = []

        if events_batch:
            flush_batch(events_batch, progress)

    def package_payload(
        self,
        model_id=None,
        dataset_id=None,
        benchmark_id=None,
        model_version=None,
        model_type=None,
        timestamp=None,
        request_id=None,
        groundtruth=None,
        prediction=None,
        confidence=None,
        features=None,
        embeddings=None,
        input_data=None,
        image_metadata=None,
        text=None,
        text_metadata=None,
        tags=None,
        audio_metadata=None,
        committed=None,
        video_metadata=None,
        validate_sample=1
    ):
        skip_validation = random.random() > validate_sample

        payload = {
            'committed': committed,
            'api_version': dioptra.__version__
        }

        if request_id:
            payload['request_id'] = request_id

        if model_id:
            payload['model_id'] = model_id

        if dataset_id:
            payload['dataset_id'] = dataset_id

        if benchmark_id:
            payload['benchmark_id'] = benchmark_id

        if model_version:
            payload['model_version'] = model_version

        if model_type:
            if isinstance(model_type, str):
                model_type = SupportedTypes[model_type]

            if validate_model_type(model_type):
                payload['model_type'] = model_type.model_type.value
                payload['input_type'] = model_type.input_type.value
            else:
                logging.warning('Invalid model_type. Skipping datapoint...')
                return []

        if timestamp:
            if isinstance(timestamp, str):
                try:
                    timestamp = dateutil.parser.isoparse(timestamp)
                except:
                    logging.warning('Could not parse string timestamp from ISO format...')

            if validate_timestamp(timestamp):
                payload['timestamp'] = timestamp.isoformat()
            else:
                logging.warning('Invalid timestamp. Replacing with current time...')
                payload['timestamp'] = datetime.datetime.utcnow().isoformat()

        if groundtruth:
            if model_type:
                if skip_validation or validate_model_type(model_type):
                    if skip_validation or validate_annotations(groundtruth, model_type):
                        payload['groundtruth'] = groundtruth
                    else:
                        logging.warning('Invalid groundtruth. Skipping datapoint...')
                        return []
                else:
                    logging.warning('Invalid model_type. Skipping datapoint...')
                    return []
            else:
                logging.warning('No model_type. Skipping datapoint...')
                return []

        if prediction:
            if model_type:
                if skip_validation or validate_model_type(model_type):
                    if skip_validation or validate_annotations(prediction, model_type):
                        payload['prediction'] = prediction
                    else:
                        logging.warning('Invalid prediction. Skipping datapoint...')
                        return []
                else:
                    logging.warning('Invalid model_type. Skipping datapoint...')
                    return []
            else:
                logging.warning('No model_type defined. Skipping datapoint...')
                return []

        if confidence:
            if skip_validation or validate_confidence(confidence):
                payload['confidence'] = confidence
            else:
                logging.warning('Invalid confidence. Skipping datapoint...')
                return []

        if features:
            if skip_validation or validate_features(features):
                payload['features'] = features
            else:
                logging.warning('Invalid features. Skipping datapoint...')
                return []

        if image_metadata:
            if skip_validation or validate_image_metadata(image_metadata):
                payload['image_metadata'] = image_metadata
            else:
                logging.warning('Invalid image_metadata. Skipping datapoint...')
                return []

        if text:
            if skip_validation or validate_text(text):
                payload['text'] = text
            else:
                logging.warning('Invalid text. Skipping datapoint...')
                return []

        if text_metadata:
            if skip_validation or validate_text_metadata(text_metadata):
                payload['text_metadata'] = text_metadata
            else:
                logging.warning('Invalid text_metadata. Skipping datapoint...')
                return []

        if embeddings:
            if skip_validation or validate_embeddings(embeddings):
                payload['embeddings'] = embeddings
            else:
                logging.warning('Invalid embeddings. Skipping datapoint...')
                return []

        if tags:
            if skip_validation or validate_tags(tags):
                payload['tags'] = tags
            else:
                logging.warning('Invalid tags. Skipping datapoint...')
                return []

        if audio_metadata:
            if skip_validation or validate_audio_metadata(audio_metadata):
                payload['audio_metadata'] = audio_metadata
            else:
                logging.warning('Invalid audio_metadata. Skipping datapoint...')
                return []

        if video_metadata:
            if skip_validation or validate_video_metadata(video_metadata):
                payload['video_metadata'] = video_metadata
            else:
                logging.warning('Invalid video_metadata. Skipping datapoint...')
                return []


        if input_data:
            if model_type:
                if skip_validation or validate_model_type(model_type):
                    if skip_validation or validate_input_data(input_data, model_type):
                        payload['input_data'] = input_data
                    else:
                        logging.warning('Invalid input_data. Skipping datapoint...')
                        return []
                else:
                    logging.warning('Invalid model_type. Skipping datapoint...')
                    return []
            else:
                logging.warning('No model model_type defined. Skipping datapoint...')
                return []

        return [payload]



import logging
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from threading import Semaphore
from concurrent.futures import ThreadPoolExecutor
from requests_futures.sessions import FuturesSession

class Client:

    def __init__(self, max_workers, queue_size, synchronous_mode):

        retry_strategy = Retry(
            total=3,
            backoff_factor=1,
            status_forcelist=[429, 500, 502, 503, 504],
            method_whitelist=['POST']
        )

        adapter = HTTPAdapter(max_retries=retry_strategy)

        if synchronous_mode:
            self.http_session = requests.Session()
        else:
            self.http_session = FuturesSession(
                executor=TaskManager(max_workers=max_workers, queue_size=queue_size))
        self.http_session.mount('https://', adapter)

        self.timeout=10
        self.synchronous_mode = synchronous_mode

    def response_hook(self, response, *args, **kwargs):
        if response.status_code != requests.codes.ok:
            logging.error('There was an error calling Dioptra API')
            response.raise_for_status()

    def call(self, payload, endpoint_url, headers):

        if self.synchronous_mode:

            self.http_session.post(
                timeout=self.timeout,
                url=endpoint_url,
                headers=headers,
                json=payload)
        else:
            self.http_session.post(
                timeout=self.timeout,
                url=endpoint_url,
                headers=headers,
                json=payload,
                hooks={
                    'response': self.response_hook
                })

class TaskManager():
    def __init__(self, max_workers, queue_size):
        self.pool = ThreadPoolExecutor(max_workers=max_workers)
        self.workers = Semaphore(max_workers + queue_size)

    def submit(self, my_method, *args, **kwargs):
        """Start a new task, blocks if queue is full."""
        self.workers.acquire()
        try:
            future = self.pool.submit(my_method, *args, **kwargs)
            future.add_done_callback(lambda x: self.workers.release())
            return future
        except:
            logging.error('Error in the task manager')
            self.workers.release()

    def shutdown(self):
        self.pool.shutdown()



import os
import pickle
import tqdm
from multiprocessing import Pool
from torch.utils.data import Dataset
import hashlib
import smart_open
from PIL import Image
from functools import partial
from dioptra.lake.utils import _decode_to_np_array

class ImageDataset(Dataset):
    def __init__(self, dataframe, transform=None):
        """
        A wrapper class on top of torch datasets.
        The goal is to allow either a data streaming from an object store or to cache the data locally and load it from there

        Parameters:
            dataframe: the datagframe to use as a data source
            transform: the transform to be applied when calling __getitem__

        """
        self.cache_dir = os.path.join(
            os.environ.get('DIOPTRA_CACHE_DIR',
            os.path.join(os.path.expanduser('~'), '.dioptra')))
        self.dataframe = dataframe
        self.transform = transform
        self.image_field = 'image_metadata.uri'
        self.segmentation_mask_field = 'groundtruth.encoded_segmentation_class_mask'
        self.use_caching = True

        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

    def __getitem__(self, index, is_prefetch=False):
        """
        Main method to get the dataset items.
        In prefetch mode, it will pull the images from an object store, skip `transform` but won't return them
        If the dataset is already prefetched, the results will be loaded from the cache dir

        Parameters:
            index: index of the datapoint to be retreived
            is_prefetch: whether in prefetch mode or not

        """
        row = self.dataframe.iloc[index].copy()

        if self.image_field in row:
            field_hash = hashlib.md5(row[self.image_field].encode()).hexdigest()
            cached_image_path = os.path.join(self.cache_dir, field_hash)
            if is_prefetch and os.path.exists(cached_image_path):
                return row
            elif self.use_caching and os.path.exists(cached_image_path):
                with open(cached_image_path, 'rb') as file:
                    row['image'] = pickle.load(file)
            else:
                img = Image.open(smart_open.open(row[self.image_field], 'rb'))
                if self.use_caching:
                    with open(cached_image_path, 'wb') as file:
                        pickle.dump(img, file)
                row['image'] = img

        if self.segmentation_mask_field in row and not is_prefetch:
            row['segmentation_class_mask'] = _decode_to_np_array(row[self.segmentation_mask_field])

        if self.transform is not None and not is_prefetch:
            return self.transform(row)

        return row


    def __len__(self):
        """
        Length of the dataset

        """
        return len(self.dataframe)

    def prefetch_images(self, num_workers=1):
        """
        Run the multi processed prefetch on the dataset
        Parameters:
            num_workers: number of processors to be used

        """
        if not self.use_caching:
            raise RuntimeError('Turn use_caching to True to be able to prefetch images')

        with Pool(num_workers) as my_pool:
            list(tqdm.tqdm(
                my_pool.imap(partial(self.__getitem__, is_prefetch=True), range(self.__len__())),
                total=self.__len__(),
                desc='Prefetching your images ...',
                ncols=100
            ))



import os
import requests

from dioptra.lake.utils import download_from_lake, _list_dataset_metadata

def list_datasets():
    """
    List all the datasets

    """

    dataset_list = []

    for metadata in _list_dataset_metadata():
        my_dataset = Dataset()
        my_dataset.dataset_id = metadata['uuid']
        my_dataset.dataset_name = metadata['display_name']
        dataset_list.append(my_dataset)

    return dataset_list

class Dataset():

    def __init__(self):

        api_key = os.environ.get('DIOPTRA_API_KEY', None)
        if api_key is None:
            raise RuntimeError('DIOPTRA_API_KEY env var is not set')

        self.api_key = api_key
        self.app_endpoint = os.environ.get('DIOPTRA_APP_ENDPOINT', 'https://app.dioptra.ai')
        self.dataset_id = None
        self.dataset_name = None

    def get_from_uuid(self, uuid):
        """
        Retreive a dataset using its uuid

        Parameters:
            uuid: uuid of the dataset
        """
        try:
            r = requests.get(f'{self.app_endpoint}/api/dataset/{uuid}', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error retreiving a dataset ...')
            raise err
        metadata = r.json()
        if len(metadata) == 0:
            raise ValueError(f'No Dataset exists for uuid {uuid}')
        self.dataset_name = metadata['display_name']
        self.dataset_id = uuid

    def get_from_name(self, name):
        """
        Retreive a dataset using its name

        Parameters:
            name: name of the dataset
        """

        for metadata in _list_dataset_metadata():
            if metadata['display_name'] == name:
                self.dataset_name = name
                self.dataset_id = metadata['uuid']
                return
        raise ValueError(f'No Dataset exists for name {name}')

    def create(self, name):
        """
        Creates a dataset. Datasets are collections of datapoints that are version controled by Dioptra

        Parameters:
            name: name of the dataset
        """

        try:
            r = requests.post(f'{self.app_endpoint}/api/dataset', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            }, json={
                'displayName': name
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error creating a dataset ...')
            raise err
        self.dataset_name = name
        self.dataset_id = r.json()['uuid']

    def commit(self, message):
        """
        Commit the changes made to the dataset that are not committed

        Parameters:
            message: message of the commit
        """

        try:
            r = requests.post(f'{self.app_endpoint}/api/dataset/{self.dataset_id}/commit', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            }, json={
                'message': message
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error committing a dataset ...')
            raise err

        # returns the commit uuid
        return r.json()['uuid']

    def history(self):
        """
        Get the dataset commit history

        """

        try:
            r = requests.get(f'{self.app_endpoint}/api/dataset/{self.dataset_id}/versions', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error fetching history for the dataset ...')
            raise err

        return r.json()

    def download_datapoints(self):
        """
        Get the datapoints in the dataset

        Parameters:
            fields: the list of fields to be downloaded.
                    By default all fields are returned, except embeddings & logits

        """

        try:
            r = requests.get(f'{self.app_endpoint}/api/dataset/{self.dataset_id}/datapoints', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error downloading a dataset ...')
            raise err
    
        return r.json()

    def download(self, fields=['*']):
        """
        Get the uuids of the datapoints in the dataset

        Parameters:
            fields: the list of fields to be downloaded.
                    By default all fields are returned, except embeddings & logits

        """

        try:
            r = requests.get(f'{self.app_endpoint}/api/dataset/{self.dataset_id}/datapoints', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error downloading a dataset ...')
            raise err

        request_ids = [entry['request_id'] for entry in r.json()]

        if len(request_ids) == 0:
            return []

        return download_from_lake(fields=fields, filters=[{
            'left': 'request_id',
            'op': 'in',
            'right': request_ids
        }])

    def add_datapoints(self, datapoint_ids):
        """
        Add datapoints to the dataset using their ids
        The datapoints will be uncommitted until a commit command is called

        Parameters:
            datapoint_ids: list of datapoint ids
        """

        try:
            r = requests.post(f'{self.app_endpoint}/api/dataset/{self.dataset_id}/add', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            }, json={
                'datapointIds': datapoint_ids
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error adding datapoints to a dataset ...')
            raise err

    def add(self, uuids):
        """
        Add datapoints to the dataset using their uuids
        The datapoints will be uncommitted until a commit command is called

        Parameters:
            uuids: list of uuids
        """

        try:
            r = requests.post(f'{self.app_endpoint}/api/datapoints/from-event-uuids', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            }, json={
                'eventUuids': uuids
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error adding datapoints to a dataset ...')
            raise err

        datapoints_ids = [datapoint['id'] for datapoint in r.json()]

        try:
            r = requests.post(f'{self.app_endpoint}/api/dataset/{self.dataset_id}/add', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            }, json={
                'datapointIds': datapoints_ids
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error adding datapoints to a dataset ...')
            raise err
    
    def remove_datapoints(self, datapoint_ids):
        """
        Remove datapoints to the dataset using their ids
        The datapoints will be uncommitted until a commit command is called unless the datapoints are aleady uncommitted

        Parameters:
            datapoint_ids: list of datapoint ids
        """

        try:
            r = requests.post(f'{self.app_endpoint}/api/dataset/{self.dataset_id}/remove', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            }, json={
                'datapointIds': datapoint_ids
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error adding datapoints to a dataset ...')
            raise err

    def remove(self, uuids):
        """
        Remove datapoints to the dataset using their uuids
        The datapoints will be uncommitted until a commit command is called unless the datapoints are aleady uncommitted

        Parameters:
            uuids: list of uuids
        """

        try:
            r = requests.post(f'{self.app_endpoint}/api/datapoints/from-event-uuids', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            }, json={
                'eventUuids': uuids
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error adding datapoints to a dataset ...')
            raise err

        datapoints_ids = [datapoint['id'] for datapoint in r.json()]

        try:
            r = requests.post(f'{self.app_endpoint}/api/dataset/{self.dataset_id}/remove', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            }, json={
                'datapointIds': datapoints_ids
            })
            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error adding datapoints to a dataset ...')
            raise err

    def checkout(self, commit_id):
        """
        Checkout the dataset as of a commit id.
        This will delete the current uncommitted changes

        Parameters:
            commit_id: id of the commit to checkout
        """

        try:
            r = requests.post(f'{self.app_endpoint}/api/dataset/{self.dataset_id}/checkout/{commit_id}', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            })

            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error checking out a dataset ...')
            raise err

        return r.json()

    def delete(self):
        """
        Delete the dataset. This is NOT reversible !!

        """

        try:
            r = requests.delete(f'{self.app_endpoint}/api/dataset/{self.dataset_id}', headers={
                'content-type': 'application/json',
                'x-api-key': self.api_key
            })

            r.raise_for_status()
        except requests.exceptions.RequestException as err:
            print('There was an error deleting a dataset ...')
            raise err


    def __str__(self):
        return f'Dataset with uuid {self.dataset_id} and name {self.dataset_name}'

    def __repr__(self):
        return self.__str__()




import os
import orjson
import mgzip
import uuid
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor

import smart_open

from dioptra.lake.utils import upload_to_lake_from_s3, wait_for_upload


class InferenceRunner():
    def __init__(self):

        api_key = os.environ.get('DIOPTRA_API_KEY', None)
        if api_key is None:
            raise RuntimeError('DIOPTRA_API_KEY env var is not set')

        s3_bucket = os.environ.get('DIOPTRA_UPLOAD_BUCKET', None)

        if s3_bucket is None:
            raise RuntimeError('DIOPTRA_UPLOAD_BUCKET env var is not set')

        s3_prefix_bucket = os.environ.get('DIOPTRA_UPLOAD_PREFIX', None)
        storage_type = os.environ.get('DIOPTRA_UPLOAD_BUCKET_TYPE', 's3')

        self.api_key = api_key
        self.s3_bucket = s3_bucket
        self.s3_prefix_bucket = s3_prefix_bucket
        self.storage_type = storage_type
        self.max_batch_size = 1000
        self.uploads = {}

    def _ingest_data(self, records):
        print('ingesting data ...')

        file_name = f'{str(uuid.uuid4())}_{datetime.utcnow().isoformat()}.ndjson.gz'

        if self.s3_prefix_bucket is not None:
            file_name = f'{self.s3_prefix_bucket}{file_name}'

        payload = b'\n'.join([orjson.dumps(record, option=orjson.OPT_SERIALIZE_NUMPY)for record in records])
        compressed_payload = mgzip.compress(payload, compresslevel=2)

        logs_url = f'{self.storage_type}://{os.path.join(self.bucket, file_name)}'

        with smart_open.open(logs_url, 'wb') as file:
            file.write(compressed_payload)

        self.uploads[file_name] = upload_to_lake_from_s3(self.s3_bucket, file_name, self.storage_type)

    def wait_for_uploads(self):
        with ThreadPoolExecutor() as executor:
            upload_ids = list(map(lambda u: u['id'], self.uploads.values()))

            return list(executor.map(wait_for_upload, upload_ids, timeout=900, chunksize=10))



from setuptools import setup, find_packages

import sys, os

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path), encoding='utf-8') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '\'' if '\'' in line else '\''
            return line.split(delim)[1]
        else:
            raise RuntimeError('Unable to find version string.')


if sys.version_info < (3, 5, 3):
    sys.exit('Sorry, Python < 3.5.3 is not supported')

long_description = read('README.md')
requirements = read('requirements.txt').splitlines()

__version__ = get_version('dioptra/__init__.py')

setup(
    name='dioptra',
    version=__version__,
    description='Client library to log data to Dioptra API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/dioptra-ai/collector-py',
    project_urls={
        'dioptra.ai': 'https://www.dioptra.ai',
    },
    author='dioptra.ai',
    author_email='info@dioptra.ai',
    license='BSD',
    python_requires='>=3.5.3',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='dioptra',
    packages=find_packages(exclude=['docs', 'tests*', 'examples']),
    include_package_data=True,
    install_requires=requirements
)
